@@ -835,6 +835,22 @@ def JzCPPMLWrapper(level=1, include_ghosts=False):
                             get_nodal_flag=_libwarpx.get_Jz_nodal_flag,
                             level=level, include_ghosts=include_ghosts)
 
+def FCPPMLWrapper(level=1, include_ghosts=False):
+    assert level>0, Exception('Coarse patch only available on levels > 0')
+    return _MultiFABWrapper(direction=None,
+                            get_lovects=_libwarpx.get_mesh_F_cp_lovects_pml,
+                            get_fabs=_libwarpx.get_mesh_F_cp_pml,
+                            get_nodal_flag=_libwarpx.get_F_nodal_flag,
+                            level=level, include_ghosts=include_ghosts)
+
+def GCPPMLWrapper(level=1, include_ghosts=False):
+    assert level>0, Exception('Coarse patch only available on levels > 0')
+    return _MultiFABWrapper(direction=None,
+                            get_lovects=_libwarpx.get_mesh_G_cp_lovects_pml,
+                            get_fabs=_libwarpx.get_mesh_G_cp_pml,
+                            get_nodal_flag=_libwarpx.get_G_nodal_flag,
+                            level=level, include_ghosts=include_ghosts)
+
 def ExFPPMLWrapper(level=0, include_ghosts=False):
     return _MultiFABWrapper(direction=0,
                             get_lovects=_libwarpx.get_mesh_electric_field_fp_lovects_pml,
@@ -897,3 +913,17 @@ def JzFPPMLWrapper(level=0, include_ghosts=False):
                             get_fabs=_libwarpx.get_mesh_current_density_fp_pml,
                             get_nodal_flag=_libwarpx.get_Jz_nodal_flag,
                             level=level, include_ghosts=include_ghosts)
+
+def FFPPMLWrapper(level=0, include_ghosts=False):
+    return _MultiFABWrapper(direction=None,
+                            get_lovects=_libwarpx.get_mesh_F_fp_lovects_pml,
+                            get_fabs=_libwarpx.get_mesh_F_fp_pml,
+                            get_nodal_flag=_libwarpx.get_F_pml_nodal_flag,
+                            level=level, include_ghosts=include_ghosts)
+
+def GFPPMLWrapper(level=0, include_ghosts=False):
+    return _MultiFABWrapper(direction=None,
+                            get_lovects=_libwarpx.get_mesh_G_fp_lovects_pml,
+                            get_fabs=_libwarpx.get_mesh_G_fp_pml,
+                            get_nodal_flag=_libwarpx.get_G_pml_nodal_flag,
+                            level=level, include_ghosts=include_ghosts)
